/***

eh.h -- Error Handler, public definitions

    Copyright (C) 1997  CSEG, Inc., All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The GNU General Public License is in the file COPYING which resides
    in the top directory of the distribution tree.

***/

#ifndef EH_H
#define EH_H 1
static const char*eh_h =
       "$Id: eh.h,v 1.2 1997/09/03 00:52:54 ccobb Exp $";

#define EH_SUPPLY_FUNCTION_NAME 1 /* program supplies function names */

/***
If the complier does not provide the
current function name in the preprocessor
definition __FUNC__, then EH_SUPPLY_FUNCTION_NAME
must be defined as 1.  This forces the programmer 
to provide explicit function names with
the EHTRACE(function_name) and EHTRY(function_name)
macros.  This is also the most portable method, because
the __FUNC__ macro is not in the C standard definition
and is therefore not universally supported.

However, if EH_SUPPLY_FUNCTION_NAME is left undefined
AND the current compiler supports __FUNC__, then the
EHTRACE() and EHTRY() macros can be used without
explicitly providing the current function's name, as
it will be picked up using __FUNC__.  This is easier
to use, but is less portable, and is therefore not
recommended.

The #if statement immediately below may use 0 or 1 as follows:

0 (recommended)     means that function names 
                    must be supplied by programmer.

1 (not recommended) may only be used when the 
                    compiler supports __FUNC__ 
                    and means that function 
                    names may be ommitted in EHTRACE() and EHTRY().

***/

#if 0 && defined(__FUNC__)
#undef EH_SUPPLY_FUNCTION_NAME /* function names from __FUNC__ */
#endif

#include <stdio.h>

#ifndef   EHENV_H
#include "ehenv.h"
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

/*** language support ***/

#if !ehSUNC
#define A(x) x
#define AV(x) x
#else
#define A(x) ()
#define AV(x) ()
#endif

/*** make sure a useable bool type exists ***/

#ifdef HAVE_CONFIG_H /*** if ./configure was used ***/

/* bool *will* be defined in ehconfig.h */

#else  /*** if configure was not used ***/

/* check the environment and act accordingly */

#if defined(__cplusplus) && ehGNUC
/* have built-in bool */
#else /* need to define bool */

/***
   When compiling on the BeBox I get different results
   depending if I use the IDE or not.  If the IDE is used
   bool seems to be defined.  This may have to do with the
   'BeHeaders' prefix defined in the IDE settings.  If the 
   command line is used bool needs to be defined here.
***/

#if ehMWCC && !defined(_SUPPORT_DEFS_H) /* Be typedefs bool in here */
typedef char bool;   /* if not, define it here */
#elif !defined(bool) /* not ehBEOS or used  _SUPPORT_DEFS_H */
#define bool char    /* define it here */
#endif
#endif /* ! defined(__cplusplus) && ehGNUC */

#ifndef false
#define false 0
#endif

#ifndef true
#define true 1
#endif

#endif /* not defined HAVE_CONFIG_H */

/*** Error Handler Definitions ***/

/* module file name definition macro: with & without ANSI __FILE__ */
#ifndef __FILE__
#define EHFILENAME(fn) static const char *__FILE__ = #fn
#else
/* we need *some* definition here to satisfy some compilers */
#define EHFILENAME(fn) typedef char _eh_dummy_ 
#endif

typedef struct
{
#define EH_SZ_PREFIX 4               /* 1-n character prefix */
   char    acPrefix[EH_SZ_PREFIX+1]; /* error prefix */
   short   hCode;                    /* error code */
} ER; /* error-handler error */

typedef struct
{
   char         *pcProgName;   /* program name */
   unsigned char ucIniState;   /* Error Handler initialization state */
   FILE         *pFilOut;      /* output that will be seen by user */
   int           iLoHi;        /* CPU byte ordering */
   ER            Er;           /* error source and number */
   unsigned long ulCtrlCIni;   /* initial EHSIMCTRLC value */
   unsigned long ulCtrlCVal;   /* incremented EHSIMCTRLC value */
   volatile unsigned char ucIntSig; /* interrupt (Ctrl-C) occurred */
   unsigned char ucNoCtrlC;    /* disable Ctrl-C checking */
   unsigned char ucNoBounds;   /* disable bounds checking */
} EH;
extern EH Eh;  /* global variables for eh library */

#define EHSTATE struct _eh_state_
EHSTATE
{
   const char   *pcFile;   /* filename of func these settings are for */
   const char   *pcFunc;   /* function these settings are for */
   int           iLine;    /* line of EHTRACE or EHTRY */
   unsigned char ucCatch;  /* current function is a catcher */
   unsigned char ucChain;  /* chaining has been requested */
   unsigned char ucCtrlC;  /* Control-C (interrupt sig) has occurred */
   unsigned char ucInhibit;/* longjmp()ing is inhibited */
   unsigned char ucShield; /* shielding has been requested */
   signed char   scError;  /* an error occurred in the current func */
   short         hLev;     /* catcher's nesting level */
   EHSTATE      *pEhSv;    /* pointer to current EHSTATE variable */
#ifndef ehEXCEPTIONHANDLING
   jmp_buf       JmpBuf;   /* longjmp target: closest catcher */
#endif
};
extern EHSTATE EhSt; /* current error handler state */

#define EHINHIBIT() (EhSt.ucInhibit=1)
#define EHSTOPCHAIN() (EhSt.ucCtrlC=EhSt.ucChain=0) /* stop chaining */
#define EHCLEAR() (Eh.Er.acPrefix[0]='\0',Eh.Er.hCode=0) /* clr error */

/***
The symbol __CHAN__ is the variable which is accessed at runtime to
determine the channel to send error messages.  
***/

/*** channel names ***/

#define EH_CHAN_UNIT  0   /* 0 = unit development */
#define EH_CHAN_ALPHA 1   /* 1 = alpha  */
#define EH_CHAN_BETA  2   /* 2 = beta */
#define EH_CHAN_MAX   2   /* highest channel number */

/*** channel setting ***/

#ifndef EH_CHAN     
#define EH_CHAN EH_CHAN_UNIT /* default to unit channel */
#else
#if EH_CHAN > EH_CHAN_MAX
#undef EH_CHAN
#define EH_CHAN EH_CHAN_MAX
#endif
#endif

#define EHCHAN(ch) static int __CHAN__ = (ch)

EHCHAN(EH_CHAN); /* define a local static variable __CHAN__ */

/*** volume levels ***/

#define EH_VOL_ERR  0 /* error messages */
#define EH_VOL_WRN  1 /* warning messages */
#define EH_VOL_REM  2 /* remark messages */
#define EH_VOL_DBG  3 /* debug messages */

/*** standard output streams ***/

#define EH_NUMCHAN       3
#define EH_BITS_PER_CHAN 5     /* total # bits per channel */
extern short *phEhChnVol;

#if 1 /* 1 = macro, 0 = function */
#define ehGetOut()\
    ( (int)(*phEhChnVol >> (EH_BITS_PER_CHAN * EH_NUMCHAN)) & 1 ) 
#else
int ehGetOut  A((void)); /* import from .c file */
#endif
bool ehSetFun A((int iChan, bool TraceOn));

#define EHSTDERR (ehGetOut() == 1 ? stdout : stderr)
#define EHCOUT (*Eh.pOStrmCout)
#define EHCERR (ehGetOut() == 1 ? cout : cerr)

/*** CTRL-C simulation ***/

#define EHCTRLC_DEFAULT ((unsigned long)~0L)
#ifdef EH_SIMCTRLC /* compiler-line request for CTRL-C simulation */
#define EHSIMCTRLC()\
 do{\
    if(Eh.ulCtrlCVal==-1L)Eh.ulCtrlCVal=Eh.ulCtrlCIni;
    if(--Eh.ulCtrlCVal==0L)ehSigInt(SIGINT);\
 }while(0)
#else
#define EHSIMCTRLC()
#endif

/* check for control-c within a function */
void ehChkCtrlC  A((void));
#define EHCTRLC() do{\
   ehChkCtrlC();\
   if (EhSt.ucCtrlC)\
   {\
      EhSt.scError = -1;\
      EHBREAK;\
   } } while(0)
#define EHCTRLC_RETURN(val) do{\
   ehChkCtrlC();\
   if (EhSt.ucCtrlC)\
   {\
      EhSt.scError = -1;\
      EHRETURN(val);\
   } } while(0)
#define EHCTRLC_LEAVE() do{\
   ehChkCtrlC();\
   if (EhSt.ucCtrlC)\
   {\
      EhSt.scError = -1;\
      EHLEAVE();\
   } } while(0)
#define EH_CTRLC 1  /* for use with ehErrVal() */

/*** miscellaneous definitions ***/

#define EH_IS_ERROR ((signed char)1)
#define EH_IS_FATAL ((signed char)-1)

/*** global function declarations ***/

void ehCaught A((const char *pcFunc));
void ehLogFun AV((const char *pcTxt, int iChn, int iVol, 
                  const char *pcFil, const char *pcFunc, int iLin,...));
void ehPError A((const char*,const char*,int,signed char,unsigned long,
                  unsigned long));

#if ehBC && (ehBCVER < 310)
#define ehErrorIs(e0,e1)  ( ehUnHash(e1) == Eh.Er )
#else
bool ehErrorIs A((unsigned long ulOffset, unsigned long ulHash));
#endif

/*** global MACRO declarations ***/

/* error (EHTHROW,EHERROR) occurred in CURRENT function */
#define ehError() (EhSt.scError != 0)

/*** EHERR, EHWRN, EHREM, EHDBG macros ***/

/** EHERR macros **/

#define EHERRX(s)\
 { s; }

#define EHERR(s)\
 ehLogFun(s, __CHAN__, EH_VOL_ERR, __FILE__, _EHFUNC_, __LINE__)

#define EHERR1(s, a1)\
 ehLogFun(s,__CHAN__, EH_VOL_ERR, __FILE__, _EHFUNC_, __LINE__,a1)

#define EHERR2(s, a1, a2)\
 ehLogFun(s,__CHAN__, EH_VOL_ERR, __FILE__, _EHFUNC_, __LINE__,a1,a2)

#define EHERR3(s, a1, a2, a3)\
 ehLogFun(s,__CHAN__, EH_VOL_ERR, __FILE__, _EHFUNC_, __LINE__,a1,a2,a3)

/** EHWRN macros **/

#define EHWRNX(s)\
 if (ehGetVol(__CHAN__) >= EH_VOL_WRN)\
 {\
   s;\
 }

#define EHWRN(s)\
 ehLogFun(s,__CHAN__, EH_VOL_WRN, __FILE__, _EHFUNC_, __LINE__)

#define EHWRN1(s, a1)\
 ehLogFun(s,__CHAN__, EH_VOL_WRN, __FILE__, _EHFUNC_, __LINE__,a1)

#define EHWRN2(s, a1, a2)\
 ehLogFun(s,__CHAN__, EH_VOL_WRN, __FILE__, _EHFUNC_, __LINE__,a1,a2)

#define EHWRN3(s, a1, a2, a3)\
 ehLogFun(s,__CHAN__, EH_VOL_WRN, __FILE__, _EHFUNC_, __LINE__,a1,a2,a3)

/** EHREM macros **/

#define EHREMX(s)\
 if (ehGetVol(__CHAN__) >= EH_VOL_REM)\
 {\
   s;\
 }

#define EHREM(s)\
 ehLogFun(s, __CHAN__, EH_VOL_REM, __FILE__, _EHFUNC_, __LINE__)

#define EHREM1(s, a1)\
 ehLogFun(s,__CHAN__, EH_VOL_REM, __FILE__, _EHFUNC_, __LINE__,a1)

#define EHREM2(s, a1, a2)\
 ehLogFun(s,__CHAN__, EH_VOL_REM, __FILE__, _EHFUNC_, __LINE__,a1,a2)

#define EHREM3(s, a1, a2, a3)\
 ehLogFun(s,__CHAN__, EH_VOL_REM, __FILE__, _EHFUNC_, __LINE__,a1,a2,a3)

/** EHDBG macros **/

#define EHDBGX(s)\
 if (ehGetVol(__CHAN__) >= EH_VOL_DBG)\
 {\
   s;\
 }

#define EHDBG(s)\
 ehLogFun(s, __CHAN__, EH_VOL_DBG, __FILE__, _EHFUNC_, __LINE__)

#define EHDBG1(s, a1)\
 ehLogFun(s,__CHAN__, EH_VOL_DBG, __FILE__, _EHFUNC_, __LINE__,a1)

#define EHDBG2(s, a1, a2)\
 ehLogFun(s,__CHAN__, EH_VOL_DBG, __FILE__, _EHFUNC_, __LINE__,a1,a2)

#define EHDBG3(s, a1, a2, a3)\
 ehLogFun(s,__CHAN__, EH_VOL_DBG, __FILE__, _EHFUNC_, __LINE__,a1,a2,a3)

#ifdef ehEXCEPTIONHANDLING

#if defined(HAVE_IOSFWD_H)
#include <iosfwd.h>
#else
class ostream;
#endif // HAVE_IOSFWD_H

ostream& ehCLog(int,int,const char *,const char *,int);

/* ehcerr macro */

#define ehcerr\
 ehCLog(__CHAN__, EH_VOL_ERR, __FILE__, _EHFUNC_, __LINE__)

/* ehcwrn macro */

#define ehcwrn\
 ehCLog(__CHAN__, EH_VOL_WRN, __FILE__, _EHFUNC_, __LINE__)

/* ehcrem macros */

#define ehcrem\
 ehCLog(__CHAN__, EH_VOL_REM, __FILE__, _EHFUNC_, __LINE__)

/* ehcdbg macro */

#define ehcdbg\
 ehCLog(__CHAN__, EH_VOL_DBG, __FILE__, _EHFUNC_, __LINE__)

#endif /* ehEXCEPTIONHANDLING */

/*** global definitions ***/

#define EHBREAK goto _EhEndCatch

/*** compile- and run-time configuration ***/

#if defined(EH_NO_MEMORY_TRACKING) || (EH_CHAN == EH_CHAN_MAX)
const bool chkOvrwrt = false; /* don't check for memory overwrites */
#else
const bool chkOvrwrt = true;  /* check for memory overwrites */
#endif

/*** THROW and ERROR ***/

#if ehBC && (ehBCVER < 310) /*** #if (ehGNUC & ehNIX) | ehBC ***/

/** special case THROW **/

#define _THROW(e0,e1)\
   ehPError(__FILE__, _EHFUNC_, __LINE__, (signed char)-1, e0,(ER)e1)

#define EHTHROW(e0,e1)\
 do{_THROW(e0,e1);EHBREAK;}while(0)

#define EHTHROW_RETURN(e0,e1,RetVal)\
 do{_THROW(e0,e1);EHRETURN(RetVal);}while(0)

#define EHTHROW_LEAVE(e0,e1)\
 do{_THROW(e0,e1);EHLEAVE();}while(0)

/** special case ERROR **/

#define _ERROR(e0,e1)\
 ehPError(__FILE__, _EHFUNC_, __LINE__, 1, e0,(ER)e1)

#define EHERROR(e0,e1)\
 do{_ERROR(e0,e1);EHBREAK;}while(0)

#define EHERROR_RETURN(e0,e1,RetVal)\
 do{_ERROR(e0,e1);EHRETURN(RetVal);}while(0)

#define EHERROR_LEAVE(e0,e1)\
 do{_ERROR(e0,e1);EHLEAVE();}while(0)

#else /** conventional definition of EHTHROW and EHERROR **/

/** conventional THROW **/

#define EHTHROW(scError)\
 do{\
    ehPError(__FILE__,_EHFUNC_,__LINE__,(signed char)-1,scError);\
    EHBREAK;\
 }while(0)
        
#define EHTHROW_RETURN(scError,RetVal)\
 do{\
    ehPError(__FILE__,_EHFUNC_,__LINE__,(signed char)-1,scError);\
    EHRETURN(RetVal);\
 }while(0)
        
#define EHTHROW_LEAVE(scError)\
 do{\
    ehPError(__FILE__,_EHFUNC_,__LINE__,(signed char)-1,scError);\
    EHLEAVE();\
 }while(0)

/** conventional ERROR **/

#define EHERROR(scError)\
 do{\
    ehPError(__FILE__,_EHFUNC_,__LINE__,1,scError);EHBREAK;\
 }while(0)
        
#define EHERROR_RETURN(scError,RetVal)\
 do{\
    ehPError(__FILE__,_EHFUNC_,__LINE__,1,scError);EHRETURN(RetVal);\
 }while(0)
        
#define EHERROR_LEAVE(scError)\
 do{\
    ehPError(__FILE__,_EHFUNC_,__LINE__,1,scError);EHLEAVE();\
 }while(0)

#endif /* conventional definitions of EHTHROW an EHERROR */

/*** function pointer types ***/

typedef int  (*pfi_pvpv)    (const void*,const void*);
typedef int  (*pfi_ppv)     (const void**);
typedef void (*pfv_pFilepv) (FILE*, const void*);
typedef void (*pfv_ulv)     (unsigned long,...); 
typedef void (*pfv_v)       (void);

/*** Resource Tracking ***/

/* internally tracked resources */

typedef enum _eh_int_res_
{
   EH_RES_UNKNOWN,  /* un-determined (yet) */
   EH_RES_C_MEM,    /* ANSI C dynamically allocated memory */
   EH_RES_CPP_MEM,  /* ANSI C++ dynamically allocated memory */
   EH_RES_FILESTR,  /* FILE stream pointer */
   EH_RES_MAX       /* number of built-in resource entries */
} EHINTRES;         /* internally tracked resource types */

#define EHNODE struct ehMinNode
EHNODE
{
   EHNODE *pEhNodeSucc;
   EHNODE *pEhNodePred;
};

typedef struct ehMinList
{
   EHNODE *pEhNodeHead;
   EHNODE *pEhNodeTail;
   EHNODE *pEhNodeTailPred;
} EHLIST;

typedef enum 
{
   EH_TOP_HVY = -1,
   EH_BALANCED,
   EH_BTM_HVY
} EHBALSTAT;       /* current balance status of a node */

#define EHAVLNODE struct _ehavlnode_
EHAVLNODE
{
   EHAVLNODE *pEhAvlNodeSelf;  /* self-referent address */
   EHAVLNODE *pEhAvlNodeUp;    /* left sub-tree */
   EHAVLNODE *pEhAvlNodeDn;    /* right sub-tree */
   EHBALSTAT  EhBalStat;       /* balance status */
   const void*pvUserData;      /* user data */
};

typedef struct
{
   EHAVLNODE*  pEhAvlNodeTop; /* top node of tree */
   pfi_pvpv    pfiCompare;    /* compare function */
   pfi_ppv     pfiDelete;     /* delete/free function */
   pfv_pFilepv pfvDump;       /* dump user data */
} EHAVL;

typedef struct
{
   bool   isInitialized; /* true if this structure was initialized */
   size_t szResCount;    /* # resources used */
   unsigned long ulResObtained; /* # times resources were obtained */
   unsigned long ulResReleased; /* # times resources were released */
   unsigned long ulResPeak;     /* # max resources used at one time */
   EHLIST EhList;    /* list header to keep track of allocation order */
   EHAVL *pEhAvlRes; /* pointer to avl tree tracking resources */
   EHAVL *pEhAvlFun; /* pointer to avl tree tracking function calls */
} EHRESTRK;

/*** EhResList.ulFlags ***/

#define EH_RES_NUMARGS_MSK 7         /* 0 to 7 arguments are allowed */
#define EH_RES_NUMARGS_WID 3         /* # bits in RES_NUMARGS_MSK */

/* resource returns no value */
#define EH_RES_VOID        (1L << (EH_RES_NUMARGS_WID + 0))
/* resource can be obtained multiple times */
#define EH_RES_ALLOW_DUPS  (1L << (EH_RES_NUMARGS_WID + 1))
/* arg 1 of obtain function must == arg 2 of release func */
#define EH_RES_A1_EQU_A2   (1L << (EH_RES_NUMARGS_WID + 2))
/* release function requires handle as arg 2: release(xxx, hndl) */
#define EH_RES_HNDL_2ND    (1L << (EH_RES_NUMARGS_WID + 3))
/* null resource handle may be passed to re-obtaining function */
#define EH_RES_ALLOW_NULL  (1L << (EH_RES_NUMARGS_WID + 4))
#define EH_RES_ALLOW_ZERO  EH_RES_ALLOW_NULL

extern EHRESTRK ehResTrk; /* internal resource tracking anchor */

void    ehAtExit      A((pfv_v pfvTermFunc));
void    ehBndSet      A((void *pvMem, size_t szMem));

bool    ehResObt      A((const char*, const char*, int, EHRESTRK*, int,
                        const char*, pfv_ulv, unsigned long,...));
bool    ehResVerify   A((EHRESTRK *pEhResTrk, unsigned long ulHndl, 
                        int iType));
void    ehResRel      A((const char*, const char*, int, EHRESTRK*,int,
                        pfv_ulv, unsigned long,...));
bool    ehResReObt    A((const char*, const char*, int, EHRESTRK*, int,
                        const char*, unsigned long, pfv_ulv,
                        unsigned long,...));
bool    ehResOnRel    A((EHRESTRK *pEhResTrk, unsigned long ulHndl1, 
                        int iType1, unsigned long ulHndl2, int iType2));

/*** support for ehgtargs.c (command-line argument processing) ***/

#define EH_GA_MAND 0x8000 /* for mandatory paramters */
typedef enum
{
   EH_GA_MESSAGE = 1, /* display given text before processing command line */
   EH_GA_USAGE,       /* description of command line usage */
   EH_GA_BOOL,        /* the mere presence of switch sets BOOL var to 1 */
   EH_GA_CBOOL,       /* the mere presence of switch sets CHAR var to 1 */
   EH_GA_HBOOL,       /* the mere presence of switch sets SHORT var to 1 */
   EH_GA_IBOOL,       /* the mere presence of switch sets INT var to 1 */
   EH_GA_CHAR,        /* place next character in char @ pvValue */
   EH_GA_ICHAR,       /* place next character in int @ pvValue */
   EH_GA_PCHAR,       /* place address of string at pointer @ pvValue */
   EH_GA_ACHAR,       /* copy text to buffer pointed to by pvValue */
   EH_GA_SHORT,       /* place next numeric string in short @ pvValue */
   EH_GA_INT,         /* place next numeric value in int @ pvValue */
   EH_GA_SIZE_T,      /* place next numeric value in size_t @ pvValue */
   EH_GA_LONG,        /* place next numeric value in long @ pvValue */
   EH_GA_FLOAT,       /* place next numeralue in float @ pvValue */
   EH_GA_DOUBLE       /* place next numeric value in double @ pvValue */
} EHGATYPE;

void  ehGtArgs  AV((volatile int * piArgc, char ***pppcArgv, ...));

/***
 ***   C++ model exception handling
 ***/

#ifdef ehEXCEPTIONHANDLING

/*** EHFUNTRACE ***/

class EHFUNTRACE
{
public:
   EHFUNTRACE();
   EHFUNTRACE(const char*pcFile, const char*pcFunc, int iLine,
         int iChan, bool chkOvrwrt);
   void ft_Args(int *piArgc, char **ppcArgv);
   void ft_Init(const char*pcFile, const char*pcFunc, int iLine,
         int iChan, bool chkOvrwrt, int * = 0, char ** = 0);
   void ft_Catch();
   void ft_LineExit(const int);
   ~EHFUNTRACE();
private:
   bool         ft_Destroyed; // work-around for compiler bug
   void         ft_PgmEnt(int *piArgc, char **ppcArgv); // program entry
   void         ft_PgmExt(); // program exit
   const char  *ft_pcFile;
   const char  *ft_pcFunc;
   int          ft_iLine;
   int          ft_iLineExit;
   int          ft_iChan;
   bool         ft_chkOvrwrt;
   EHSTATE     ft_EhSt;
};

/*** EHMAIN ***/

#ifndef _EHFUNC_
#define EHMAIN(ac,av)\
 Eh.pcProgName = av[0];\
 const char *volatile _EHFUNC_ = "main";\
 EHFUNTRACE ehFunTrace;\
 do {\
 ehFunTrace.ft_Catch();\
 try {\
 ehFunTrace.ft_Init(\
       __FILE__,_EHFUNC_,__LINE__,__CHAN__,chkOvrwrt, &ac, av);
#else /* _EHFUNC_ is defined (as NULL) */
#define EHMAIN(ac,av)\
 Eh.pcProgName = av[0];\
 EHFUNTRACE ehFunTrace;\
 do {\
 ehFunTrace.ft_Catch();\
 try {\
 ehFunTrace.ft_Init(\
       __FILE__,_EHFUNC_,__LINE__,__CHAN__,chkOvrwrt,&ac,av);
#endif /* _EHFUNC_ */

/*** EHENDMAIN ***/

#define EHENDMAIN\
 }catch(...){ehCaught(_EHFUNC_);}}while(0);_EhEndCatch:EHINHIBIT()

/*** EHDONE ***/

#define EHDONE(val)\
 do{ehFunTrace.ft_LineExit(__LINE__);return(val);}while(0)

/*** EHTRACE ***/

#ifndef _EHFUNC_
#ifdef EH_SUPPLY_FUNCTION_NAME
#define EHTRACE(sFunc)\
 EHSIMCTRLC();\
 const char *volatile _EHFUNC_ = #sFunc;\
 EHFUNTRACE ehFunTrace(\
       __FILE__,_EHFUNC_,(const int)__LINE__,__CHAN__,chkOvrwrt)
#else /* don't need to EH_SUPPLY_FUNCTION_NAME */
#define EHTRACE()\
 EHSIMCTRLC();\
 const char *volatile _EHFUNC_ = __FUNC__;\
 EHFUNTRACE ehFunTrace(\
       __FILE__,_EHFUNC_,(const int)__LINE__,__CHAN__,chkOvrwrt)
#endif /* don't need to EH_SUPPLY_FUNCTION_NAME */
#else /* _EHFUNC_ is defined (as NULL -- for no-function-tracing) */
#ifdef EH_SUPPLY_FUNCTION_NAME
#define EHTRACE(sFunc)\
 EHSIMCTRLC();\
 EHFUNTRACE ehFunTrace(\
       __FILE__,#sFunc,(const int)__LINE__,__CHAN__,chkOvrwrt)
#else /* don't need to EH_SUPPLY_FUNCTION_NAME */
#define EHTRACE()\
 EHSIMCTRLC();\
 EHFUNTRACE ehFunTrace(\
       __FILE__,__FUNC__,(const int)__LINE__,__CHAN__,chkOvrwrt)
#endif /* don't need to EH_SUPPLY_FUNCTION_NAME */
#endif /* _EHFUNC_ is defined (as NULL) */

#define EHLEAVE()\
 do{ehFunTrace.ft_LineExit(__LINE__);return;}while(0)

#define EHRETURN(val)\
 do{ehFunTrace.ft_LineExit(__LINE__);return(val);}while(0)

#ifdef EH_SUPPLY_FUNCTION_NAME
#define EHTRY(sFunc)\
 EHTRACE(sFunc);\
 ehFunTrace.ft_Catch();\
 do { try
#else /* don't need to EH_SUPPLY_FUNCTION_NAME */
#define EHTRY()\
 EHTRACE();\
 ehFunTrace.ft_Catch();\
 do { try
#endif /* don't need to EH_SUPPLY_FUNCTION_NAME */

/*** EHCATCH ***/

#define EHCATCH\
 catch(...){ehCaught(_EHFUNC_);}}while(0);_EhEndCatch:EHINHIBIT()

/***
 ***   C implementation of exception handling
 ***/

#else  /* !ehEXCEPTIONHANDLING */

/* coming soon! */

#endif /* ehEXCEPTIONHANDLING */

#endif /* EH_H */
